@@ -24,7 +24,7 @@ from persistent.tests.utils import skipIfNoCExtension
 
 _marker = object()
 
-class DummyPersistent(object):
+class DummyPersistent:
     _Persistent__ring = None
 
     def _p_invalidate(self):
@@ -40,15 +40,25 @@ class DummyPersistent(object):
         self._p_state = UPTODATE
 
 
-class DummyConnection(object):
+class DummyConnection:
     pass
 
 
+class ClosedConnection(DummyConnection):
+    def __init__(self, test):
+        self.test = test
+
+    def setstate(self, obj): # pragma: no cover
+        self.test.fail("Connection is closed")
+
+    def register(self, obj):
+        """Does nothing."""
+
 def _len(seq):
     return len(list(seq))
 
 
-class PickleCacheTestMixin(object):
+class PickleCacheTestMixin:
 
     # py2/3 compat
     assertRaisesRegex = getattr(unittest.TestCase,
@@ -204,7 +214,7 @@ class PickleCacheTestMixin(object):
     def test___setitem___persistent_class(self):
 
         KEY = b'pclass'
-        class pclass(object):
+        class pclass:
             _p_oid = KEY
             _p_jar = DummyConnection()
         cache = self._makeOne(pclass._p_jar)
@@ -237,7 +247,7 @@ class PickleCacheTestMixin(object):
 
         KEY = b'pclass'
         cache = self._makeOne()
-        class pclass(object):
+        class pclass:
             _p_oid = KEY
             _p_jar = DummyConnection()
         cache = self._makeOne()
@@ -301,10 +311,7 @@ class PickleCacheTestMixin(object):
 
 
     def _numbered_oid(self, i):
-        # Python 3.4 doesn't support % on bytes,
-        # so we go the long way
-        oid_s = 'oid_%04d' % i
-        return oid_s.encode('ascii')
+        return b'oid_%04d' % i
 
     def _populate_cache(self, cache, count=100,
                         state_0=UPTODATE,
@@ -381,6 +388,84 @@ class PickleCacheTestMixin(object):
         for oid in oids:
             self.assertTrue(cache.get(oid) is None)
 
+    def test_full_sweep_clears_weakrefs_in_interface(self, sweep_method='full_sweep'):
+        # https://github.com/zopefoundation/persistent/issues/149
+        # Sweeping the cache clears weak refs (for PyPy especially)
+        # In the real world, this shows up in the interaction with
+        # persistent objects and zope.interface/zope.component,
+        # so we use an Interface to demonstrate. This helps find issues
+        # like needing to run GC more than once, etc, because of how the
+        # object is referenced.
+        from zope.interface import Interface
+
+        gc.disable()
+        self.addCleanup(gc.enable)
+
+        jar = ClosedConnection(self)
+        cache = self._makeOne(jar, 0)
+
+        # Make a persistent object, put it in the cache as saved
+        class P(self._getRealPersistentClass()):
+            "A real persistent object that can be weak referenced"
+
+        p = P()
+        p._p_jar = jar
+        p._p_oid = b'\x01' * 8
+        cache[p._p_oid] = p
+        p._p_changed = False
+
+        # Now, take a weak reference to it from somewhere far away.
+        Interface.subscribe(p)
+
+        # Remove the original object
+        del p
+
+        # Sweep the cache.
+        getattr(cache, sweep_method)()
+
+        # Now, try to use that weak reference. If the weak reference is
+        # still around, this will raise the error about the connection
+        # being closed.
+        Interface.changed(None)
+
+    def test_incrgc_clears_weakrefs_in_interface(self):
+        self.test_full_sweep_clears_weakrefs_in_interface(sweep_method='incrgc')
+
+    def test_full_sweep_clears_weakrefs(self, sweep_method='incrgc'):
+        # like test_full_sweep_clears_weakrefs_in_interface,
+        # but directly using a weakref. This is the simplest version of the test.
+        from weakref import ref as WeakRef
+        gc.disable()
+        self.addCleanup(gc.enable)
+
+        jar = ClosedConnection(self)
+        cache = self._makeOne(jar, 0)
+
+        # Make a persistent object, put it in the cache as saved
+        class P(self._getRealPersistentClass()):
+            """A real persistent object that can be weak referenced."""
+
+        p = P()
+        p._p_jar = jar
+        p._p_oid = b'\x01' * 8
+        cache[p._p_oid] = p
+        p._p_changed = False
+
+        # Now, take a weak reference to it
+        ref = WeakRef(p)
+
+        # Remove the original object
+        del p
+
+        # Sweep the cache.
+        getattr(cache, sweep_method)()
+
+        # Now, try to use that weak reference; it should be gone.
+        p = ref()
+        self.assertIsNone(p)
+
+    def test_incrgc_clears_weakrefs(self):
+        self.test_full_sweep_clears_weakrefs(sweep_method='incrgc')
 
     def test_minimize(self):
         cache = self._makeOne()
@@ -469,7 +554,7 @@ class PickleCacheTestMixin(object):
 
     def test_new_ghost_w_pclass_non_ghost(self):
         KEY = b'123'
-        class Pclass(object):
+        class Pclass:
             _p_oid = None
             _p_jar = None
         cache = self._makeOne()
@@ -480,7 +565,7 @@ class PickleCacheTestMixin(object):
 
     def test_new_ghost_w_pclass_ghost(self):
         KEY = b'123'
-        class Pclass(object):
+        class Pclass:
             _p_oid = None
             _p_jar = None
         cache = self._makeOne()
@@ -534,7 +619,7 @@ class PickleCacheTestMixin(object):
 
     def test_debug_info_w_persistent_class(self):
         KEY = b'pclass'
-        class pclass(object):
+        class pclass:
             _p_oid = KEY
             _p_jar = DummyConnection()
         cache = self._makeOne(pclass._p_jar)
@@ -634,7 +719,7 @@ class PickleCacheTestMixin(object):
 
     def test_invalidate_persistent_class_calls_p_invalidate(self):
         KEY = b'pclass'
-        class pclass(object):
+        class pclass:
             _p_oid = KEY
             _p_jar = DummyConnection()
             invalidated = False
@@ -743,24 +828,24 @@ class PythonPickleCacheTests(PickleCacheTestMixin, unittest.TestCase):
         self.assertFalse(updated)
 
     def test___delitem___w_persistent_class(self):
-        cache, key = super(PythonPickleCacheTests, self).test___delitem___w_persistent_class()
+        cache, key = super().test___delitem___w_persistent_class()
         self.assertNotIn(key, cache.persistent_classes)
 
     def test___setitem___ghost(self):
-        cache = super(PythonPickleCacheTests, self).test___setitem___ghost()
+        cache = super().test___setitem___ghost()
         self.assertEqual(cache.ringlen(), 0)
 
     def test___setitem___persistent_class(self):
-        cache = super(PythonPickleCacheTests, self).test___setitem___persistent_class()
+        cache = super().test___setitem___persistent_class()
         self.assertEqual(_len(cache.items()), 0)
 
     def test_new_ghost_w_pclass_non_ghost(self):
-        cache, Pclass, key = super(PythonPickleCacheTests, self).test_new_ghost_w_pclass_non_ghost()
+        cache, Pclass, key = super().test_new_ghost_w_pclass_non_ghost()
         self.assertEqual(Pclass._p_jar, cache.jar)
         self.assertIs(cache.persistent_classes[key], Pclass)
 
     def test_new_ghost_w_pclass_ghost(self):
-        cache, Pclass, key = super(PythonPickleCacheTests, self).test_new_ghost_w_pclass_ghost()
+        cache, Pclass, key = super().test_new_ghost_w_pclass_ghost()
         self.assertEqual(Pclass._p_jar, cache.jar)
         self.assertIs(cache.persistent_classes[key], Pclass)
 
@@ -867,7 +952,7 @@ class PythonPickleCacheTests(PickleCacheTestMixin, unittest.TestCase):
 
     def test_invalidate_hit_pclass(self):
         KEY = b'123'
-        class Pclass(object):
+        class Pclass:
             _p_oid = KEY
             _p_jar = DummyConnection()
         cache = self._makeOne(Pclass._p_jar)
@@ -877,16 +962,16 @@ class PythonPickleCacheTests(PickleCacheTestMixin, unittest.TestCase):
         self.assertNotIn(KEY, cache.persistent_classes)
 
     def test_debug_info_w_normal_object(self):
-        obj, info = super(PythonPickleCacheTests, self).test_debug_info_w_normal_object()
+        obj, info = super().test_debug_info_w_normal_object()
         self.assertEqual(info[1], len(gc.get_referents(obj)))
         self.assertEqual(info[3], UPTODATE)
 
     def test_debug_info_w_ghost(self):
-        ghost, info = super(PythonPickleCacheTests, self).test_debug_info_w_ghost()
+        ghost, info = super().test_debug_info_w_ghost()
         self.assertEqual(info[1], len(gc.get_referents(ghost)))
 
     def test_debug_info_w_persistent_class(self):
-        pclass, info = super(PythonPickleCacheTests, self).test_debug_info_w_persistent_class()
+        pclass, info = super().test_debug_info_w_persistent_class()
         self.assertEqual(info[3], UPTODATE)
         self.assertEqual(info[1], len(gc.get_referents(pclass)))
 
@@ -925,7 +1010,7 @@ class PythonPickleCacheTests(PickleCacheTestMixin, unittest.TestCase):
     def test_init_with_cacheless_jar(self):
         # Sometimes ZODB tests pass objects that don't
         # have a _cache
-        class Jar(object):
+        class Jar:
             was_set = False
             def __setattr__(self, name, value):
                 if name == '_cache':
@@ -1049,7 +1134,7 @@ class PythonPickleCacheTests(PickleCacheTestMixin, unittest.TestCase):
 
 
     def test_new_ghost_obj_already_in_cache(self):
-        base_result = super(PythonPickleCacheTests, self).test_new_ghost_obj_already_in_cache()
+        base_result = super().test_new_ghost_obj_already_in_cache()
         cache, key, candidate = base_result
         # If we're sneaky and remove the OID and jar, then we get the duplicate
         # key error. Removing them only works because we're not using a real
@@ -1102,6 +1187,44 @@ class PythonPickleCacheTests(PickleCacheTestMixin, unittest.TestCase):
         self.assertEqual(cache.cache_non_ghost_count, 0)
         self.assertEqual(len(cache), 0)
 
+    def test_interpreter_finalization_ffi_cleanup(self):
+        # When the interpreter is busy garbage collecting old objects
+        # and clearing their __dict__ in random orders, the CFFI cleanup
+        # ``ffi.gc()`` cleanup hooks we use on CPython don't
+        # raise errors.
+        #
+        # Prior to Python 3.8, when ``sys.unraisablehook`` was added,
+        # the only way to know if this test fails is to look for AttributeError
+        # on stderr.
+        #
+        # But wait, it gets worse. Prior to https://foss.heptapod.net/pypy/cffi/-/issues/492
+        # (CFFI > 1.14.5, unreleased at this writing), CFFI ignores
+        # ``sys.unraisablehook``, so even on 3.8 the only way to know
+        # a failure is to watch stderr.
+        #
+        # See https://github.com/zopefoundation/persistent/issues/150
+
+        import sys
+        unraised = []
+        try:
+            old_hook = sys.unraisablehook
+        except AttributeError:
+            pass
+        else: # pragma: no cover
+            sys.unraisablehook = unraised.append
+            self.addCleanup(setattr, sys, 'unraisablehook', old_hook)
+
+        cache = self._makeOne()
+        oid = self._numbered_oid(42)
+        o = cache[oid] = self._makePersist(oid=oid)
+        # Clear the dict, or at least part of it.
+        # This is coupled to ``cleanup_hook``
+        if cache.data.cleanup_hook:
+            del cache.data._addr_to_oid
+        del cache[oid]
+
+        self.assertEqual(unraised, [])
+
 
 @skipIfNoCExtension
 class CPickleCacheTests(PickleCacheTestMixin, unittest.TestCase):
@@ -1134,12 +1257,12 @@ class CPickleCacheTests(PickleCacheTestMixin, unittest.TestCase):
             verifyObject(IExtendedPickleCache, self._makeOne(), tentative=True)
 
     def test___setitem___persistent_class(self):
-        cache = super(CPickleCacheTests, self).test___setitem___persistent_class()
+        cache = super().test___setitem___persistent_class()
         self.assertEqual(_len(cache.items()), 1)
 
     def test_cache_garbage_collection_bytes_with_cache_size_0(self):
 
-        class DummyConnection(object):
+        class DummyConnection:
             def register(self, obj):
                 pass
 
@@ -1182,6 +1305,30 @@ class CPickleCacheTests(PickleCacheTestMixin, unittest.TestCase):
         self.assertEqual(len(cache), 0)
 
 
+class TestWeakValueDictionary(unittest.TestCase):
+
+    def _getTargetClass(self):
+        from persistent.picklecache import _WeakValueDictionary
+        return _WeakValueDictionary
+
+    def _makeOne(self):
+        return self._getTargetClass()()
+
+    @unittest.skipIf(PYPY, "PyPy doesn't have the cleanup_hook")
+    def test_cleanup_hook_gc(self):
+        # A more targeted test than ``test_interpreter_finalization_ffi_cleanup``
+        # See https://github.com/zopefoundation/persistent/issues/150
+        wvd = self._makeOne()
+
+        class cdata:
+            o = object()
+            pobj_id = id(o)
+        wvd['key'] = cdata.o
+
+        wvd.__dict__.clear()
+        wvd.cleanup_hook(cdata)
+
+
 def test_suite():
     return unittest.defaultTestLoader.loadTestsFromName(__name__)
 
